@@ -2,6 +2,7 @@ from abc import ABC
 from typing import Any
 from pipe import traverse, map
 
+from plotly.subplots import make_subplots
 import plotly.graph_objects as go
 from plotly.basedatatypes import BaseTraceType
 
@@ -30,12 +31,16 @@ class TraceWrapper(ABC):
   def mk_trace(self, data: list[dict[str, Any]] | dict[str, Any]) -> BaseTraceType:
     fpath_data = {}
     for key, fpath in self.fpaths.items():
-      item = fpath.extract_data(data)
+      item = fpath._extract_data(data)
       if type(item) == list and len(item) == 1:
         fpath_data[key] = item[0]
       else:
         fpath_data[key] = item
 
+    # print(f'mk_trace: {fpath_data}')
+    # for key, item in fpath_data.items():
+    #   print(f'{key}: {len(item)} datapoints')
+
     return self.graph_object(
       **(fpath_data | self.args)
     )
@@ -71,6 +76,11 @@ class Contour(TraceWrapper):
   graph_object = go.Contour
 
 
+class Table(TraceWrapper):
+  """See https://plotly.com/python/contour-plots/"""
+  graph_object = go.Table
+
+
 # Distributions
 class Box(TraceWrapper):
   """See https://plotly.com/python/box-plots/"""
@@ -253,4 +263,8 @@ class Figure:
       self.data = self.subgrounds.execute(self.req)
 
       for trace in self.traces:
-        self.figure.add_trace(trace.mk_trace(self.data))
\ No newline at end of file
+        self.figure.add_trace(trace.mk_trace(self.data))
+  
+  # @staticmethod
+  # def mk_subplots(rows, cols, **kwargs):
+  #   return make_subplots(rows, cols, **kwargs)
\ No newline at end of file
